# $Id: Mysql.pm 1.1 Tue, 30 Sep 1997 01:28:08 +0200 joe $

package Bundle::Mysql;

$VERSION = '0.01';

1;

__END__

=head1 NAME

Bundle::Mysql - A bundle to install MysqlPerl, DBD::mysql and prerequisites

=head1 SYNOPSIS

C<perl -MCPAN -e 'install Bundle::Mysql'>

=head1 CONTENTS

DBI

Data::ShowTable

Mysql

DBD::mysql

=head1 DESCRIPTION

This bundle includes all that's needed to run both DBD::mysql and MysqlPerl.

